@@ -15,6 +15,10 @@
 #include <signal.h>       /* for signal()    */
 #include <stdlib.h>       /* for abort()     */
 
+#ifdef __linux__
+#include <sys/prctl.h>    /* for prctl() */
+#endif /* __linux__ */
+
 void
 gu_abort (void)
 {
@@ -22,6 +26,12 @@ gu_abort (void)
     struct rlimit core_limits = { 0, 0 };
     setrlimit (RLIMIT_CORE, &core_limits);
 
+#ifdef __linux__
+    /* Linux with its coredump piping option requires additional care.
+     * See e.g. https://patchwork.kernel.org/patch/1091782/ */
+    prctl(PR_SET_DUMPABLE, 0, 0, 0, 0);
+#endif /* __linux__ */
+
     /* restore default SIGABRT handler */
     signal (SIGABRT, SIG_DFL);
 
