@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2010-2024 Codership Oy <info@codership.com>
+ * Copyright (C) 2010-2016 Codership Oy <info@codership.com>
  */
 
 #include "gcache_rb_store.hpp"
@@ -14,7 +14,6 @@
 #include <gu_hash.h>
 
 #include <cassert>
-#include <iostream> // std::cerr
 
 namespace gcache
 {
@@ -28,16 +27,6 @@ namespace gcache
     {
         write_preamble(false);
 
-        for (seqno2ptr_iter_t i = seqno2ptr_.begin(); i != seqno2ptr_.end();)
-        {
-            if (ptr2BH(*i)->ctx == BH_ctx_t(this)) {
-                i = seqno2ptr_.erase(i);
-            }
-            else {
-                ++i;
-            }
-        }
-
         first_ = start_;
         next_  = start_;
 
@@ -54,15 +43,12 @@ namespace gcache
     void
     RingBuffer::constructor_common() {}
 
-    RingBuffer::RingBuffer (ProgressCallback*  pcb,
-                            const std::string& name,
+    RingBuffer::RingBuffer (const std::string& name,
                             size_t             size,
                             seqno2ptr_t&       seqno2ptr,
                             gu::UUID&          gid,
-                            int const          dbg,
                             bool const         recover)
     :
-        pcb_       (pcb),
         fd_        (name, check_size(size)),
         mmap_      (fd_),
         preamble_  (static_cast<char*>(mmap_.ptr)),
@@ -73,17 +59,14 @@ namespace gcache
         next_      (first_),
         seqno2ptr_ (seqno2ptr),
         gid_       (gid),
-        seqno_locked_(SEQNO_MAX),
         size_cache_(end_ - start_ - sizeof(BufferHeader)),
         size_free_ (size_cache_),
         size_used_ (0),
         size_trail_(0),
 //        mallocs_   (0),
 //        reallocs_  (0),
-        debug_     (dbg & DEBUG),
         open_      (true)
     {
-        assert((uintptr_t(start_) % MemOps::ALIGNMENT) == 0);
         constructor_common ();
         open_preamble(recover);
         BH_clear (BH_cast(next_));
@@ -115,31 +98,26 @@ namespace gcache
     {
         for (seqno2ptr_t::iterator i(i_begin); i != i_end;)
         {
-            seqno2ptr_t::iterator j(i);
+            seqno2ptr_t::iterator j(i); ++i;
+            BufferHeader* const bh (ptr2BH (j->second));
 
-            /* advance i to next set element skipping holes */
-            do { ++i; } while ( i != i_end && !*i);
-
-            BufferHeader* const bh(ptr2BH(*j));
-
-            if (gu_likely (BH_is_released(bh) && bh->seqno_g < seqno_locked_))
+            if (gu_likely (BH_is_released(bh)))
             {
                 seqno2ptr_.erase (j);
+                empty_buffer(bh);
 
                 switch (bh->store)
                 {
-                case BUFFER_IN_RB:
-                    discard(bh);
-                    break;
+                case BUFFER_IN_RB:  discard(bh); break;
                 case BUFFER_IN_MEM:
                 {
-                    MemStore* const ms(static_cast<MemStore*>(BH_ctx(bh)));
+                    MemStore* const ms(static_cast<MemStore*>(bh->ctx));
                     ms->discard(bh);
                     break;
                 }
                 case BUFFER_IN_PAGE:
                 {
-                    Page*      const page (static_cast<Page*>(BH_ctx(bh)));
+                    Page*      const page (static_cast<Page*>(bh->ctx));
                     PageStore* const ps   (PageStore::page_store(page));
                     ps->discard(bh);
                     break;
@@ -162,7 +140,6 @@ namespace gcache
     BufferHeader*
     RingBuffer::get_new_buffer (size_type const size)
     {
-        assert((size % MemOps::ALIGNMENT) == 0);
         assert_size_free();
 
         BH_assert_clear(BH_cast(next_));
@@ -255,7 +232,6 @@ namespace gcache
 #endif
 
     found_space:
-        assert((uintptr_t(ret) % MemOps::ALIGNMENT) == 0);
         size_used_ += size;
         assert (size_used_ <= size_cache_);
         assert (size_free_ >= size);
@@ -264,11 +240,12 @@ namespace gcache
         BufferHeader* const bh(BH_cast(ret));
         bh->size    = size;
         bh->seqno_g = SEQNO_NONE;
+        bh->seqno_d = SEQNO_ILL;
         bh->flags   = 0;
         bh->store   = BUFFER_IN_RB;
-        bh->ctx     = reinterpret_cast<BH_ctx_t>(this);
+        bh->ctx     = this;
+
         next_ = ret + size;
-        assert((uintptr_t(next_) % MemOps::ALIGNMENT) == 0);
         assert (next_ + sizeof(BufferHeader) <= end_);
         BH_clear (BH_cast(next_));
         assert_sizes();
@@ -413,25 +390,28 @@ namespace gcache
 
         if (size_cache_ == size_free_) return;
 
-        /* Invalidate seqnos for all ordered buffers (so that they can't be
-         * recovered on restart. Also find the last seqno'd RB buffer. */
+        /* Find the last seqno'd RB buffer. It is likely to be close to the
+         * end of released buffers chain. */
         BufferHeader* bh(0);
 
-        for (seqno2ptr_t::iterator i(seqno2ptr_.begin());
-             i != seqno2ptr_.end(); ++i)
+        for (seqno2ptr_t::reverse_iterator r(seqno2ptr_.rbegin());
+             r != seqno2ptr_.rend(); ++r)
         {
-            BufferHeader* const b(ptr2BH(*i));
+            BufferHeader* const b(ptr2BH(r->second));
             if (BUFFER_IN_RB == b->store)
             {
 #ifndef NDEBUG
                 if (!BH_is_released(b))
                 {
-                    log_fatal << "Buffer " << b << " is not released.";
+                    log_fatal << "Buffer "
+                              << reinterpret_cast<const void*>(r->second)
+                              << ", seqno_g " << b->seqno_g << ", seqno_d "
+                              << b->seqno_d << " is not released.";
                     assert(0);
                 }
 #endif
-                b->seqno_g = SEQNO_NONE;
                 bh = b;
+                break;
             }
         }
 
@@ -536,10 +516,10 @@ namespace gcache
     void
     RingBuffer::print (std::ostream& os) const
     {
-        os  << "this: " << static_cast<const void*>(this)
-            << "\nstart_ : " << BH_cast(start_)
-            << "\nfirst_ : " << BH_cast(first_) << ", off: " << first_ - start_
-            << "\nnext_  : " << BH_cast(next_) << ", off: " << next_  - start_
+        os  << "\nstart_ : " << reinterpret_cast<void*>(start_)
+            << "\nend_   : " << reinterpret_cast<void*>(end_)
+            << "\nfirst  : " << first_ - start_
+            << "\nnext   : " << next_  - start_
             << "\nsize   : " << size_cache_
             << "\nfree   : " << size_free_
             << "\nused   : " << size_used_;
@@ -555,6 +535,7 @@ namespace gcache
     void
     RingBuffer::write_preamble(bool const synced)
     {
+        static int const VERSION(1);
         uint8_t* const preamble(reinterpret_cast<uint8_t*>(preamble_));
 
         std::ostringstream os;
@@ -567,10 +548,10 @@ namespace gcache
             if (!seqno2ptr_.empty())
             {
                 os << PR_KEY_SEQNO_MIN << ' '
-                   << seqno2ptr_.index_front() << '\n';
+                   << seqno2ptr_.begin()->first << '\n';
 
                 os << PR_KEY_SEQNO_MAX << ' '
-                   << seqno2ptr_.index_back() << '\n';
+                   << seqno2ptr_.rbegin()->first << '\n';
 
                 os << PR_KEY_OFFSET << ' ' << first_ - preamble << '\n';
             }
@@ -597,8 +578,8 @@ namespace gcache
     void
     RingBuffer::open_preamble(bool const do_recover)
     {
-        int version(0); // used only for recovery on upgrade
         uint8_t* const preamble(reinterpret_cast<uint8_t*>(preamble_));
+        int version(0);
         long long seqno_max(SEQNO_ILL);
         long long seqno_min(SEQNO_ILL);
         off_t offset(-1);
@@ -635,32 +616,23 @@ namespace gcache
            version = 0;
         }
 
-        if (offset < -1 ||
-            (preamble + offset + sizeof(BufferHeader)) > end_ ||
-            (version >= 2 && offset >= 0 && (offset % MemOps::ALIGNMENT)))
+        if (offset < -1 || (preamble + offset + sizeof(BufferHeader)) > end_)
         {
            log_warn << "Bogus offset in GCache ring buffer preamble: "
                     << offset << ". Assuming unknown.";
            offset = -1;
         }
 
-        log_info << "GCache DEBUG: opened preamble:"
-                 << "\nVersion: " << version
-                 << "\nUUID: " << gid_
-                 << "\nSeqno: " << seqno_min << " - " << seqno_max
-                 << "\nOffset: " << offset
-                 << "\nSynced: " << synced;
-
         if (do_recover)
         {
             if (gid_ != gu::UUID())
             {
-                log_info << "Recovering GCache ring buffer: version: " <<version
+                log_info << "Recovering GCache ring buffer: version: " << version
                          << ", UUID: " << gid_ << ", offset: " << offset;
 
                 try
                 {
-                    recover(offset - (start_ - preamble), version);
+                    recover(offset - (start_ - preamble));
                 }
                 catch (gu::Exception& e)
                 {
@@ -685,42 +657,21 @@ namespace gcache
         write_preamble(true);
     }
 
-    /* Helper callback class to specilize for different progress types */
-    template <typename T>
-    class recover_progress_callback : public gu::Progress<T>::Callback
-    {
-    public:
-        recover_progress_callback(gcache::ProgressCallback* pcb)
-            : pcb_(pcb)
-        {}
-        ~recover_progress_callback() {}
-        void operator()(T const total, T const done)
-        {
-            if (pcb_) (*pcb_)(total, done);
-        }
-    private:
-        recover_progress_callback(const recover_progress_callback&);
-        recover_progress_callback& operator=(recover_progress_callback);
-        ProgressCallback* pcb_;
-    };
-
-    seqno_t
-    RingBuffer::scan(off_t const offset, int const scan_step)
+    int64_t
+    RingBuffer::scan(off_t const offset)
     {
         int segment_scans(0);
-        seqno_t seqno_max(SEQNO_ILL);
+        int64_t seqno_max(-1);
         uint8_t* ptr;
         BufferHeader* bh;
         size_t collision_count(0);
-        seqno_t erase_up_to(-1);
+        int64_t erase_up_to(-1);
         uint8_t* segment_start(start_);
         uint8_t* segment_end(end_ - sizeof(BufferHeader));
 
         /* start at offset (first segment) if we know it and it is valid */
         if (offset >= 0)
         {
-            assert(0 == (offset % scan_step));
-
             if (start_ + offset + sizeof(BufferHeader) < segment_end)
                 /* we know exaclty where the first segment starts */
                 segment_start = start_ + offset;
@@ -729,10 +680,8 @@ namespace gcache
                 segment_scans = 1;
         }
 
-        recover_progress_callback<ptrdiff_t> scan_progress_callback(pcb_);
-        gu::Progress<ptrdiff_t> progress(&scan_progress_callback,
-                                         "GCache::RingBuffer initial scan",
-                                         " bytes", end_ - start_, 1<<22/*4Mb*/);
+        gu::Progress<ptrdiff_t> progress("GCache::RingBuffer initial scan",
+                                         " bytes", end_ - start_, 1<<22 /*4Mb*/);
 
         while (segment_scans < 2)
         {
@@ -746,35 +695,30 @@ namespace gcache
              ptr + bh->size <= segment_end &&                   \
              BH_test(BH_cast(ptr + bh->size)))
 
-#define GCACHE_SCAN_ADVANCE(amount)             \
-            ptr += amount;                      \
-            progress.update(amount);            \
-            bh = BH_cast(ptr);
-
-
             while (GCACHE_SCAN_BUFFER_TEST)
             {
-                assert((uintptr_t(bh) % scan_step) == 0);
-
                 bh->flags |= BUFFER_RELEASED;
-                bh->ctx    = uint64_t(this);
+                bh->ctx    = this;
 
-                seqno_t const seqno_g(bh->seqno_g);
+                int64_t const seqno_g(bh->seqno_g);
 
                 if (gu_likely(seqno_g > 0))
                 {
-                    bool const collision(
-                        seqno_g <= seqno_max &&
-                        seqno_g >= seqno2ptr_.index_begin() &&
-                        seqno2ptr_[seqno_g] != seqno2ptr_t::null_value());
+                    const std::pair<seqno2ptr_iter_t, bool>& res(
+                        seqno_g > seqno_max ? seqno_max = seqno_g,
+                        std::pair<seqno2ptr_iter_t, bool>(
+                            seqno2ptr_.insert(seqno2ptr_.end(),
+                                              seqno2ptr_pair_t(seqno_g, bh + 1)),
+                            true) :
+                        seqno2ptr_.insert(seqno2ptr_pair_t(seqno_g, bh + 1))
+                        );
 
-                    if (gu_unlikely(collision))
+                    if (gu_unlikely (false == res.second))
                     {
                         collision_count++;
 
                         /* compare two buffers */
-                        seqno2ptr_t::const_reference old_ptr
-                            (seqno2ptr_[seqno_g]);
+                        const void* const old_ptr(res.first->second);
                         BufferHeader* const old_bh
                             (old_ptr ? ptr2BH(old_ptr) : NULL);
 
@@ -802,11 +746,10 @@ namespace gcache
                                                        sizeof(cs_old)));
                         std::ostringstream msg;
 
-                        msg << "Attempt (" << collision_count
-                            << ") to reuse the same seqno: " << seqno_g
+                        msg << "Attempt to reuse the same seqno: " << seqno_g
                             << ". New ptr = " << new_ptr << ", " << bh
                             << ", cs: " << gu::Hexdump(cs_new, sizeof(cs_new))
-                            << ", previous ptr = " << old_ptr;
+                            << ", previous ptr = " << res.first->second;
 
                         empty_buffer(bh); // this buffer is unusable
                         assert(BH_is_released(bh));
@@ -820,8 +763,11 @@ namespace gcache
                             {
                                 empty_buffer(old_bh);
                                 assert(BH_is_released(old_bh));
+                                res.first->second = NULL;
+                                /* mark entry as invalid, but don't remove from
+                                 * the map to block this seqno */
 
-                                if (erase_up_to < seqno_g) erase_up_to =seqno_g;
+                                if (erase_up_to < seqno_g) erase_up_to = seqno_g;
                             }
                         }
 
@@ -831,58 +777,28 @@ namespace gcache
                             log_info << "Contents are the same, discarding "
                                      << new_ptr;
                         } else {
+                            assert(NULL == res.first->second);
                             log_info << "Contents differ. Discarding both.";
                         }
                     }
-                    else
-                    {
-                        try
-                        {
-                            seqno2ptr_.insert(seqno_g, bh + 1);
-                        }
-                        catch (std::exception& e)
-                        {
-                            seqno_t const sb(seqno2ptr_.empty() ? SEQNO_ILL :
-                                             seqno2ptr_.index_begin());
-                            seqno_t const se(seqno2ptr_.empty() ? SEQNO_ILL :
-                                             seqno2ptr_.index_end());
-                            log_warn << "Exception while mapping writeset "
-                                     << bh << " into [" << sb << ", " << se
-                                     << "): '" << e.what()
-                                     << "'. Aborting GCache recovery.";
-                            /* Buffer scanning was interrupted ungracefully -
-                             * this means that we failed to recover the most
-                             * recent writesets. As such anything that was
-                             * potentially recovered before is useless.
-                             * This will cause full cache reset in recover() */
-                            seqno2ptr_.clear(SEQNO_ILL);
-
-                            BH_clear(bh); // to pacify assert() below
-                            next_ = ptr;
-                            goto out;
-                        }
-                        seqno_max = std::max(seqno_g, seqno_max);
-                    }
                 }
 
-                GCACHE_SCAN_ADVANCE(bh->size);
+                progress.update(bh->size);
+                ptr += bh->size;
+                bh = BH_cast(ptr);
             }
 
             if (!BH_is_clear(bh))
             {
-                if (start_ == segment_start && ptr != first_ &&
-                    ptr + bh->size != first_)
-                    /* ptr + bh->size == first_ means that there is only one
-                     * segment starting at first_ and the space between start_
-                     * and first_ occupied by discarded buffers. */
+                if (start_ == segment_start && ptr != first_)
                 {
                     log_warn << "Failed to scan the last segment to the end. "
                             "Last events may be missing. Last recovered event: "
                              << gid_ << ':' << seqno_max;
                 }
 
-                /* end of segment, close it */
-                BH_clear(bh);
+                /* end of file, do best effort */
+                if (end_ - sizeof(BufferHeader) == segment_end) BH_clear(bh);
             }
 
             if (offset > 0 && segment_start == start_ + offset)
@@ -891,8 +807,7 @@ namespace gcache
                 assert(1 == segment_scans);
                 first_ = segment_start;
                 size_trail_ = end_ - ptr;
-                // there must be at least one buffer header between the segments
-                segment_end = segment_start - sizeof(BufferHeader);
+                segment_end = segment_start;
                 segment_start = start_;
             }
             else if (offset < 0 && segment_start == start_)
@@ -900,21 +815,22 @@ namespace gcache
                 /* started with the second segment, try to find the first one */
                 assert(1 == segment_scans);
                 next_ = ptr;
-                GCACHE_SCAN_ADVANCE(sizeof(BufferHeader));
 
-                while (ptr + sizeof(BufferHeader) < end_ &&
-                       !GCACHE_SCAN_BUFFER_TEST)
+                while (!GCACHE_SCAN_BUFFER_TEST &&
+                       ptr + sizeof(BufferHeader) < end_)
                 {
-                    GCACHE_SCAN_ADVANCE(scan_step);
+                    progress.update(1);
+                    ptr += 1;
+                    bh = BH_cast(ptr);
+                    /* it would have been nice to use alignment of 8 or so */
                 }
 
                 if (GCACHE_SCAN_BUFFER_TEST)
                 {
-                    /* looks like a valid buffer, a beginning of a segment */
                     segment_start = ptr;
                     first_ = segment_start;
                 }
-                else if (ptr + sizeof(BufferHeader) >= end_)
+                else if (ptr + sizeof(BufferHeader) == end_)
                 {
                     /* perhaps it was a single segment starting at start_ */
                     first_ = start_;
@@ -954,48 +870,20 @@ namespace gcache
                 }
             }
 #undef GCACHE_SCAN_BUFFER_TEST
-#undef GCACHE_SCAN_ADVANCE
         } // while (segment_scans < 2)
-    out:
-        assert(BH_is_clear(BH_cast(next_)));
-        progress.finish();
 
-        if (debug_)
-        {
-            log_info
-                << "RB: end of scan(): seqno2ptr: "
-                << seqno2ptr_.index_begin() << " - " << seqno2ptr_.index_end()
-                << ", seqno_max: " << seqno_max;
-            log_info << "RB: " << *this;
-            dump_map();
-        }
+        progress.finish();
 
         return erase_up_to;
     }
 
-    static bool assert_ptr_seqno(seqno2ptr_t& map,
-                                 const void* const ptr,
-                                 seqno_t     const seqno)
-    {
-        const BufferHeader* const bh(ptr2BH(ptr));
-        if (bh->seqno_g != seqno)
-        {
-            assert(0);
-            map.clear(SEQNO_NONE);
-            return true;
-        }
-        return false;
-    }
-
     void
-    RingBuffer::recover(off_t const offset, int version)
+    RingBuffer::recover(off_t const offset)
     {
-        static const char* const diag_prefix ="Recovering GCache ring buffer: ";
+        static const char* const diag_prefix = "Recovering GCache ring buffer: ";
 
         /* scan the buffer and populate seqno2ptr map */
-        seqno_t const lowest(scan(offset, version > 0 ? MemOps::ALIGNMENT : 1)
-                             + 1);
-        /* lowest is the lowest valid seqno based on collisions during scan */
+        int64_t const lower(scan(offset));
 
         if (!seqno2ptr_.empty())
         {
@@ -1004,46 +892,54 @@ namespace gcache
 
             /* find the last gapless seqno sequence */
             seqno2ptr_t::reverse_iterator r(seqno2ptr_.rbegin());
-            assert(*r);
-            seqno_t const seqno_max(seqno2ptr_.index_back());
-            seqno_t       seqno_min(seqno2ptr_.index_front());
+            seqno_t const seqno_max(r->first);
+            seqno_t       seqno_min(seqno2ptr_.begin()->first);
 
-            /* need to search for seqno gaps */
-            assert(seqno_max >= lowest);
-            if (lowest == seqno_max)
+            if (lower > 0
+                /* collisions detected */ ||
+                seqno2ptr_t::size_type(seqno_max - seqno_min + 1) >
+                seqno2ptr_.size()
+                /* not all seqnos present */)
             {
-                seqno2ptr_.clear(SEQNO_NONE);
-                goto full_reset;
+                /* need to search for seqno gaps */
+                assert(seqno_max >= lower);
+                if (lower == seqno_max)
+                {
+                    seqno2ptr_.clear();
+                    goto full_reset;
+                }
+
+                seqno_min = seqno_max;
+                ++r;
+                for (; r != seqno2ptr_.rend() && r->first > lower; ++r)
+                {
+                    assert(r->second != NULL);
+
+                    if (r->first + 1 == seqno_min)
+                        seqno_min = r->first;
+                    else
+                        break;
+                }
             }
-
-            seqno_min = seqno_max;
-            if (assert_ptr_seqno(seqno2ptr_, *r, seqno_min)) goto full_reset;
-
-            /* At this point r and seqno_min both point at the last element in
-             * the map. Scan downwards and bail out on the first hole.*/
-            ++r;
-            for (; r != seqno2ptr_.rend() && *r && seqno_min > lowest; ++r)
+            else
             {
-                --seqno_min;
-                if (assert_ptr_seqno(seqno2ptr_, *r,seqno_min)) goto full_reset;
+                r = seqno2ptr_.rend();
             }
-            /* At this point r points to one below seqno_min */
 
             log_info << diag_prefix << "found gapless sequence " << seqno_min
                      << '-' << seqno_max;
 
             if (r != seqno2ptr_.rend())
             {
-                assert(seqno_min > seqno2ptr_.index_begin());
                 log_info << diag_prefix << "discarding seqnos "
-                         << seqno2ptr_.index_begin() << '-' << seqno_min - 1;
+                         << seqno2ptr_.begin()->first << '-' << r->first;
 
                 /* clear up seqno2ptr map */
                 for (; r != seqno2ptr_.rend(); ++r)
                 {
-                    if (*r) empty_buffer(ptr2BH(*r));
+                    if (r->second) empty_buffer(ptr2BH(r->second));
                 }
-                seqno2ptr_.erase(seqno2ptr_.begin(),seqno2ptr_.find(seqno_min));
+                seqno2ptr_.erase(seqno2ptr_.begin(), seqno2ptr_.find(seqno_min));
             }
             assert(seqno2ptr_.size() > 0);
 
@@ -1057,30 +953,19 @@ namespace gcache
 
                 bh = BH_next(bh);
 
-                if (gu_unlikely(0 == bh->size)) bh = BH_cast(start_);// rollover
+                if (gu_unlikely(0 == bh->size)) bh = BH_cast(start_); // rollover
             }
             first_ = reinterpret_cast<uint8_t*>(bh);
 
             /* trim next_: start with the last seqno and scan forward up to the
              * current next_. Update to the end of the last non-empty buffer. */
-            bh = ptr2BH(seqno2ptr_.back());
+            bh = ptr2BH(seqno2ptr_[seqno_max]);
             BufferHeader* last_bh(bh);
             while (bh != BH_cast(next_))
             {
                 if (gu_likely(bh->size) > 0)
                 {
-                    bool const inconsistency(
-                        BH_next(bh) > BH_cast(end_ - sizeof(BufferHeader)) ||
-                        bh->ctx != BH_ctx_t(this)
-                        );
-
-                    if (gu_unlikely(inconsistency))
-                    {
-                        assert(0);
-                        log_warn << diag_prefix << "Corrupt buffer leak1: "<<bh;
-                        goto full_reset;
-                    }
-
+                    assert(bh->seqno_d >= SEQNO_ILL);
                     assert(bh->size > sizeof(BufferHeader));
 
                     if (bh->seqno_g > 0) last_bh = bh;
@@ -1094,22 +979,6 @@ namespace gcache
             }
             next_ = reinterpret_cast<uint8_t*>(BH_next(last_bh));
 
-            /* Even if previous buffers were not aligned, make sure from
-             * now on they are - adjust next_ pointer and last buffer size */
-            if (uintptr_t(next_) % MemOps::ALIGNMENT)
-            {
-                uint8_t* const n(MemOps::align_ptr(next_));
-                assert(n > next_);
-                size_type const size_diff(n - next_);
-                assert(size_diff < MemOps::ALIGNMENT);
-                assert(last_bh->size > 0);
-                last_bh->size += size_diff;
-                next_ = n;
-                assert(BH_next(last_bh) == BH_cast(next_));
-            }
-            assert((uintptr_t(next_) % MemOps::ALIGNMENT) == 0);
-            BH_clear(BH_cast(next_));
-
             /* at this point we must have at least one seqno'd buffer */
             assert(next_ != first_);
 
@@ -1120,202 +989,49 @@ namespace gcache
             estimate_space();
 
             /* now discard all the locked-in buffers (see seqno_reset()) */
+            gu::Progress<size_t> progress(
+                "GCache::RingBuffer unused buffers scan",
+                " bytes", size_used_, 1<<22 /* 4Mb */);
+
             size_t total(0);
             size_t locked(0);
-
+            bh = BH_cast(first_);
+            while (bh != BH_cast(next_))
             {
-                recover_progress_callback<size_t>
-                    unused_progress_callback(pcb_);
-                gu::Progress<size_t> progress(
-                    &unused_progress_callback,
-                    "GCache::RingBuffer unused buffers scan",
-                    " bytes", size_used_, 1<<22 /* 4Mb */);
-
-                bh = BH_cast(first_);
-                while (bh != BH_cast(next_))
+                if (gu_likely(bh->size > 0))
                 {
-                    if (gu_likely(bh->size > 0))
+                    total++;
+                    if (gu_unlikely(SEQNO_ILL == bh->seqno_g))
                     {
-                        bool const inconsistency(
-                            BH_next(bh) > BH_cast(end_ - sizeof(BufferHeader))
-                            ||
-                            bh->ctx != BH_ctx_t(this)
-                            );
-
-                        if (gu_unlikely(inconsistency))
-                        {
-                            assert(0);
-                            log_warn << diag_prefix << "Corrupt buffer leak2: "
-                                     << bh;
-                            goto full_reset;
-                        }
-
-                        total++;
-
-                        if (gu_likely(bh->seqno_g > 0))
-                        {
-                            free(bh); // on recovery no buffer is used
-                        }
-                        else
-                        {
-                            /* anything that is not ordered must be discarded */
-                            assert(SEQNO_NONE == bh->seqno_g ||
-                                   SEQNO_ILL  == bh->seqno_g);
-                            locked++;
-                            empty_buffer(bh);
-                            discard(bh);
-                            size_used_ -= bh->size;
-                            // size_free_ is taken care of in discard()
-                        }
-
-                        bh = BH_next(bh);
-                    }
-                    else
-                    {
-                        bh = BH_cast(start_); // rollover
+                        locked++;
+                        discard(bh);
                     }
 
-                    progress.update(bh->size);
+                    bh = BH_next(bh);
+                }
+                else
+                {
+                     bh = BH_cast(start_); // rollover
                 }
 
-                progress.finish();
+                progress.update(bh->size);
             }
 
-            /* No buffers on recovery should be in used state */
-            assert(0 == size_used_);
+            progress.finish();
 
-            log_info << diag_prefix << "found "
+            log_info << "GCache DEBUG: RingBuffer::recover(): found "
                      << locked << '/' << total << " locked buffers";
-            log_info << diag_prefix << "free space: "
-                     << size_free_ << '/' << size_cache_;
+            log_info << "GCache DEBUG: RingBuffer::recover(): used space: "
+                     << size_used_ << '/' << size_cache_;
 
             assert_sizes();
-
-            if (debug_)
-            {
-                log_info << *this;
-                dump_map();
-            }
         }
         else
         {
         full_reset:
-            log_info << diag_prefix <<"Recovery failed, need to do full reset.";
+            log_info << diag_prefix << "didn't recover any events.";
             reset();
         }
     }
 
-    static void
-    print_chain(const uint8_t* const rb_start, const uint8_t* const chain_start,
-                const uint8_t* const chain_end, size_t const count,
-                const char* const type_str)
-    {
-        ptrdiff_t const start_off(chain_start - rb_start);
-        ptrdiff_t const end_off(chain_end - rb_start);
-        std::cerr <<  start_off << "\t" << end_off << "\t"
-                  << end_off - start_off << "\t" << count << "\t"
-                  << type_str << std::endl;
-    }
-
-    void
-    RingBuffer::dump_map() const
-    {
-        enum chain_t
-        {
-            ORDERED,
-            UNORDERED,
-            RELEASED,
-            NONE
-        };
-
-        static const char* chain_str[] =
-            { "ORDERED", "UNORDERED", "RELEASED", "NONE" };
-
-        size_t chain_size[] = { 0, 0, 0, 0 };
-        size_t chain_count[] = { 0, 0, 0, 0 };
-
-        chain_t chain(NONE);
-        const uint8_t* chain_start(start_);
-        size_t count;
-
-        bool next(false);
-        const uint8_t* ptr(start_);
-        const BufferHeader* bh(BH_const_cast(ptr));
-
-        log_info << "RB start_";
-        log_info << bh;
-        for (int i(0); i < 2; i++)
-        {
-            while (!BH_is_clear(bh))
-            {
-                if (first_ == ptr && i == 0)
-                {
-                    goto first; // rare situation when there is only
-                    // one segment in the start/middle
-                }
-
-                size_t const offset(bh->size);
-                chain_t const typ(bh->seqno_g >= 0 ? ORDERED : UNORDERED);
-                if (chain != typ)
-                { // new chain starts
-                    if (chain != NONE)
-                    { // old chain ends
-                        print_chain(start_, chain_start, ptr, count,
-                                    chain_str[chain]);
-                        chain_count[chain] += count;
-                    }
-                    chain = typ;
-                    chain_start = ptr;
-                    count = 0;
-                }
-                count++;
-                chain_size[typ] += offset;
-                chain_size[RELEASED] += offset * BH_is_released(bh);
-                chain_count[RELEASED] += BH_is_released(bh);
-
-                ptr += offset;
-                bh = BH_const_cast(ptr);
-            }
-            // old chain ends
-            print_chain(start_, chain_start, ptr, count, chain_str[chain]);
-            chain_count[chain] += count;
-            if (1 == i) break; // final segment read
-
-            log_info << "RB next_";
-            log_info << bh << ", off: " << ptr - start_;
-            next = true;
-
-            log_info << "RB middle gap: " << first_ - ptr;
-
-            ptr = first_;
-            bh = BH_const_cast(ptr);
-        first:
-            chain = NONE;
-            count = 0;
-            log_info << "RB first_";
-            log_info << bh << ", off: " << ptr - start_;
-        }
-
-        if (!next)
-        {
-            log_info << "RB next_";
-        }
-        else
-        {
-            log_info << "RB rollover";
-        }
-        log_info << bh << ", off: " << ptr - start_;
-        log_info << "RB trailing space: " << end_ - ptr;
-
-        log_info << "RB space usage:"
-                 << "\nORDERED  : " << chain_size[ORDERED]
-                 << "\nUNORDERED: " << chain_size[UNORDERED]
-                 << "\nRELEASED : " << chain_size[RELEASED]
-                 << "\nNONE     : " << chain_size[NONE];
-        log_info << "RB buf counts:"
-                 << "\nORDERED  : " << chain_count[ORDERED]
-                 << "\nUNORDERED: " << chain_count[UNORDERED]
-                 << "\nRELEASED : " << chain_count[RELEASED]
-                 << "\nNONE     : " << chain_count[NONE];
-    }
 } /* namespace gcache */
