@@ -539,7 +539,7 @@ void gcomm::GMCast::handle_connected(Proto* rp)
     log_debug << "transport " << tp << " connected";
     if (rp->state() == Proto::S_INIT)
     {
-        log_debug << "sending hanshake";
+        log_debug << "sending handshake";
         // accepted socket was waiting for underlying transport
         // handshake to finish
         rp->send_handshake();
@@ -615,7 +615,7 @@ void gcomm::GMCast::handle_established(Proto* est)
         i = remote_addrs_.insert_unique(
             make_pair(est->remote_addr(),
                       AddrEntry(gu::datetime::Date::now(),
-                                gu::datetime::Date::max(),
+                                gu::datetime::Date::now(),
                                 est->remote_uuid())));
     }
 
@@ -979,7 +979,6 @@ void gcomm::GMCast::reconnect()
         }
     }
 
-
     for (i = remote_addrs_.begin(); i != remote_addrs_.end(); i = i_next)
     {
         i_next = i, ++i_next;
@@ -1319,7 +1318,7 @@ void gcomm::GMCast::handle_up(const void*        id,
             return;
         }
 
-        if (msg.type() >= Message::T_USER_BASE)
+        if (msg.type() >= Message::GMCAST_T_USER_BASE)
         {
             gu_trace(send_up(Datagram(dg, dg.offset() + msg.serial_size()),
                              ProtoUpMeta(msg.source_uuid())));
@@ -1361,7 +1360,7 @@ void gcomm::GMCast::handle_up(const void*        id,
                 return;
             }
 
-            if (msg.type() >= Message::T_USER_BASE)
+            if (msg.type() >= Message::GMCAST_T_USER_BASE)
             {
                 if (evict_list().empty() == false &&
                     evict_list().find(msg.source_uuid()) != evict_list().end())
@@ -1446,7 +1445,7 @@ void gcomm::GMCast::handle_up(const void*        id,
 
 int gcomm::GMCast::handle_down(Datagram& dg, const ProtoDownMeta& dm)
 {
-    Message msg(version_, Message::T_USER_BASE, uuid(), 1, segment_);
+    Message msg(version_, Message::GMCAST_T_USER_BASE, uuid(), 1, segment_);
 
     // handle relay set first, skip these peers below
     if (relay_set_.empty() == false)
