@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2008-2025 Codership Oy <info@codership.com>
+ * Copyright (C) 2008-2016 Codership Oy <info@codership.com>
  *
  * $Id$
  */
@@ -22,7 +22,7 @@
 #define GCS_STATE_FREP       0x01 // group representative
 #define GCS_STATE_FCLA       0x02 // count last applied (for JOINED node)
 #define GCS_STATE_FBOOTSTRAP 0x04 // part of prim bootstrap process
-#define GCS_STATE_FSTATELESS 0x08 // arbitrator or otherwise stateless node
+#define GCS_STATE_ARBITRATOR 0x08 // arbitrator or otherwise incomplete node
 
 #ifdef GCS_STATE_MSG_ACCESS
 typedef struct gcs_state_msg
@@ -33,21 +33,14 @@ typedef struct gcs_state_msg
     gcs_seqno_t      prim_seqno;    // last PC state seqno
     gcs_seqno_t      received;      // last action seqno (received up to)
     gcs_seqno_t      cached;        // earliest action cached
-    gcs_seqno_t      last_applied;  // last applied action reported by node
-    gcs_seqno_t      vote_seqno;    // last seqno node voted on
-    int64_t          vote_res;      // the vote reported by node
     const char*      name;          // human assigned node name
     const char*      inc_addr;      // incoming address string
     int              version;       // version of state message
     int              gcs_proto_ver;
     int              repl_proto_ver;
     int              appl_proto_ver;
-    int              prim_gcs_ver;
-    int              prim_repl_ver;
-    int              prim_appl_ver;
     int              prim_joined;   // number of joined nodes in its last PC
     int              desync_count;
-    uint8_t          vote_policy;   // voting policy the node is using
     gcs_node_state_t prim_state;    // state of the node in its last PC
     gcs_node_state_t current_state; // current state of the node
     uint8_t          flags;
@@ -63,25 +56,20 @@ typedef struct gcs_state_quorum
     gu_uuid_t   group_uuid;   //! group UUID
     gcs_seqno_t act_id;       //! next global seqno
     gcs_seqno_t conf_id;      //! configuration id
-    gcs_seqno_t last_applied; //! group-wide commit cut
     bool        primary;      //! primary configuration or not
     int         version;      //! state excahnge version (max understood by all)
     int         gcs_proto_ver;
     int         repl_proto_ver;
     int         appl_proto_ver;
-    uint8_t     vote_policy;
 }
 gcs_state_quorum_t;
 
-#define GCS_VOTE_ZERO_WINS 1
-
 #define GCS_QUORUM_NON_PRIMARY (gcs_state_quorum_t){    \
         GU_UUID_NIL,                                    \
         GCS_SEQNO_ILL,                                  \
         GCS_SEQNO_ILL,                                  \
-        GCS_SEQNO_ILL,                                  \
         false,                                          \
-        -1, -1, -1, -1, GCS_VOTE_ZERO_WINS              \
+        -1, -1, -1, -1                                  \
     }
 
 extern gcs_state_msg_t*
@@ -91,10 +79,6 @@ gcs_state_msg_create (const gu_uuid_t* state_uuid,
                       gcs_seqno_t      prim_seqno,
                       gcs_seqno_t      received,
                       gcs_seqno_t      cached,
-                      gcs_seqno_t      last_applied,
-                      gcs_seqno_t      vote_seqno,
-                      int64_t          vote_res,
-                      uint8_t          vote_policy,
                       int              prim_joined,
                       gcs_node_state_t prim_state,
                       gcs_node_state_t current_state,
@@ -103,9 +87,6 @@ gcs_state_msg_create (const gu_uuid_t* state_uuid,
                       int              gcs_proto_ver,
                       int              repl_proto_ver,
                       int              appl_proto_ver,
-                      int              prim_gcs_ver,
-                      int              prim_repl_ver,
-                      int              prim_appl_ver,
                       int              desync_count,
                       uint8_t          flags);
 
@@ -160,19 +141,6 @@ gcs_state_msg_name (const gcs_state_msg_t* state);
 extern const char*
 gcs_state_msg_inc_addr (const gcs_state_msg_t* state);
 
-/* Get last applied action seqno */
-gcs_seqno_t
-gcs_state_msg_last_applied (const gcs_state_msg_t* state);
-
-/* Get last vote */
-void
-gcs_state_msg_last_vote (const gcs_state_msg_t* state,
-                         gcs_seqno_t& seqno, int64_t& res);
-
-/* Get vote policy */
-uint8_t
-gcs_state_msg_vote_policy (const gcs_state_msg_t* state);
-
 /* Get supported protocols */
 extern void
 gcs_state_msg_get_proto_ver (const gcs_state_msg_t* state,
@@ -197,7 +165,7 @@ gcs_state_msg_flags (const gcs_state_msg_t* state);
  *         quorum parameter */
 extern long
 gcs_state_msg_get_quorum (const gcs_state_msg_t* states[],
-                          size_t                 states_num,
+                          long                   states_num,
                           gcs_state_quorum_t*    quorum);
 
 /* Print state message contents to buffer */
